from os.path import join
from icecube import icetray, dataclasses
from icecube.icetray import I3Units

# written by Markus Voge <voge@physik.uni-bonn.de>

class Domain(object):

    SPTS = 1
    SPS = 2
    WISC = 3
    DESY = 4
    UNKOWN = 5

def getDomain():
    import socket
    host = socket.gethostname()
    domain = host[host.find(".")+1:]

    if 'sps' in domain:
        return Domain.SPS
    elif 'spts' in domain:
        return Domain.SPTS
    elif domain == "icecube.wisc.edu":
        return Domain.WISC
    elif domain == 'ifh.de':
        return Domain.DESY

class PulseMaskShortcutter(icetray.I3ConditionalModule):
    '''
    Create a copy of a pulse mask, pointing directly to a specific source,
    leaving out all the intermediate masks.  This module loops through the mask
    sources for the given PulseMaskName until a non-mask source is reached,
    then it creates a copy of the PulseMaskName pointing directly to the
    non-mask source, without intermediate steps, put into the frame as
    ShortcutName.  Alternatively, one can name a desired source with
    OrigSourceName.
    '''
    def __init__(self, context):
        icetray.I3ConditionalModule.__init__(self, context)
        self.AddParameter("PulseMaskName", "Name of the I3RecoPulseSeriesMapMask to create a shortcut for", "")
        self.AddParameter("ShortcutName", "Name of the shortcut to put into the frame", "")
        self.AddParameter("OrigSourceName", "Name of the desired source the shortcut should point to. Scans for first non-mask source if not set.", "")
        self.AddOutBox("OutBox")

    def Configure(self):
        self.pulseMaskName = self.GetParameter("PulseMaskName")
        self.shortcutName = self.GetParameter("ShortcutName")
        self.origSourceName = self.GetParameter("OrigSourceName")

    def DAQ(self, frame):
        self.PushFrame(frame)

    def Physics(self, frame):
        mask = frame[self.pulseMaskName]
        sourceName = mask.source
        if frame.Has(self.origSourceName):
            if sourceName == self.origSourceName:
                ##print "Pulse mask is already shortcut. Putting copy of it into the frame."
                frame.Put(self.shortcutName, mask) # since mask is already pointing directly to desired source
            else:
                sourceName = self.origSourceName
        else: # Loop until you find something rock solid (no mask anymore)
            sourceMask = frame[sourceName]
            while type(sourceMask) == dataclasses.I3RecoPulseSeriesMapMask:
                sourceName = sourceMask.source
                sourceMask = frame[sourceName]
        ##print "Found original source pulses named \'%s\', creating shortcut." % sourceName
        shortcutMask = dataclasses.I3RecoPulseSeriesMapMask(frame, sourceName, mask.apply(frame)) # create a shortcut pulse mask
        ##print "Putting shortcut pulse mask into the frame."
        frame.Put(self.shortcutName, shortcutMask)
        self.PushFrame(frame)

    def Finish(self):
        return True


@icetray.traysegment
def SplineReco(tray, name, pulses="CleanedMuonPulses", SplineDir=None,
        ParametrizationName="Para", MinimizerName="Minuit", PathToCramerRaoTable="/path/to/table",
        If = lambda f: True):
    '''
    Perform a spline-reco reconstruction using photospline tables in SplineDir.
    '''
    icetray.load("spline-reco", False)
    # path to bare muon splines
    domain = getDomain()
    #if not SplineDir:
    #    if domain == Domain.SPS or domain == Domain.SPTS:
    #        SplineDir = '/mnt/data/pnflocal/InfBareMu_mie_spline/'
    #    elif domain == Domain.WISC:
    #        SplineDir = '/net/user/schatto/InfBareMu_mie_spline/'
    #    elif domain == Domain.DESY:
    #        SplineDir = '/lustre/fs6/group/i3/mvoge/photospline/InfBareMu_mie_spline/'
    #    else:
    #        SplineDir = "/opt/photorec2/"
    
    timingSpline = join(SplineDir, 'InfBareMu_mie_prob_z20a10.fits')
    amplitudeSpline = join(SplineDir, 'InfBareMu_mie_abs_z20a10.fits')

    ### next services provide spline access

    # bare muon spline with 4 ns convolution
    tray.AddService("I3PhotoSplineServiceFactory", name+"_splines4",
         AmplitudeTable = amplitudeSpline, # path to amplitude spline (abs)
         TimingTable    = timingSpline, # path to timing spline (prob)
         TimingSigma    = 4, # pdf convolution in ns
         ServiceName    = name+'_SplineService4', # name of the service
         )

    # seed service without vertex time shift, PandelMPE as first guess
    # used for the SplineMPE fit
    tray.AddService( "I3BasicSeedServiceFactory", name+"_MPESeedNoShift",
            FirstGuesses =                [name+"_BestFit"],
            ChargeFraction =              0.9,                      # Default
            FixedEnergy =                 float( "nan" ),           # Default
            MaxMeanTimeResidual =         1000.0*I3Units.ns,        # Default
            NChEnergyGuessPolynomial =    [],                       # Default
            SpeedPolice =                 True,                     # Default
            AddAlternatives =             "None",                   # Default
            OnlyAlternatives =            False,                    # Default
            )

    # the spline reco likelihood doing a default MPE reco --> faster and worse than with all modifications
    tray.AddService("I3SplineRecoLikelihoodFactory", name+"_SplineMPEllh",
           PhotonicsService = name+"_SplineService4", # spline services
           Pulses = pulses,
           Likelihood = "MPE",
           NoiseRate = 10*I3Units.hertz,
           )

    tray.AddModule("I3SimpleFitter", name+"_SplineMPE",
            SeedService = name+"_MPESeedNoShift",
            Parametrization = ParametrizationName,
            LogLikelihood = name+"_SplineMPEllh",
            Minimizer = MinimizerName,
            NonStdName =  "Params",                 # Default
            StoragePolicy =  "OnlyBestFit",         # Default
            If = If
            )

    tray.AddModule("CramerRao", name + "_CramerRaoSplineMPE",
            InputResponse = pulses,
            InputTrack = name+"_SplineMPE",
            OutputResult = name+"_CramerRao_SplineMPE",
            AllHits = True,
            DoubleOutput = True,
            PathToTable = PathToCramerRaoTable,
            z_dependent_scatter = True,
            If = If
            )


@icetray.traysegment
def AngularError(tray, name, pulses="CleanedMuonPulses", MinimizerName="Minuit", MPEllhName="MPEllh",
        If = lambda f: True):
    '''
    Perform more advanced angular error estimation with Paraboloid.
    MinimizerName should be a name of an already installed minimizer service.
    MPEllhName should be a name of an already installed Pandel MPE likelihood service.
    '''
    # this is the first time we use mpe as a seed -> install the service
    tray.AddService("I3BasicSeedServiceFactory", name+"_BestFitSeed",
            FirstGuesses =                 [ name+"_BestFit" ],      # This is the best available fit that succeeded
            ChargeFraction =               0.9,                      # Default
            FixedEnergy =                  float( "nan" ),           # Default
            MaxMeanTimeResidual =          1000.0*I3Units.ns,        # Default
            NChEnergyGuessPolynomial =     [],                       # Default
            SpeedPolice =                  True,                     # Default
            AddAlternatives =              "None",                   # Default
            OnlyAlternatives =             False                     # Default
            )

    tray.AddService("I3BasicSeedServiceFactory", name+"_GridpointVertexCorrection",
            InputReadout =                 pulses,
            FirstGuesses =                 [ name+"_BestFit" ],      # This is the best available fit that succeeded
            TimeShiftType =                "TFirst",                 # ! Use TFirst for vertex correction
            ChargeFraction =               0.9,                      # Default
            FixedEnergy =                  float( "nan" ),           # Default
            MaxMeanTimeResidual =          1000.0*I3Units.ns,        # Default
            NChEnergyGuessPolynomial =     [],                       # Default
            SpeedPolice =                  True,                     # Default
            AddAlternatives =              "None",                   # Default
            AltTimeShiftType =             "TFirst",                 # Default
            OnlyAlternatives =             False                     # Default
            )

    icetray.load("paraboloid", False)
    tray.AddModule( "I3ParaboloidFitter", name+"_BestFit_Paraboloid",
            SeedService =                   name+"_BestFitSeed",        # seed service from above
            LogLikelihood =                 MPEllhName,            # likelihood service has been installed earlier
            MaxMissingGridPoints =          1,                      # ! Allow 1 missfit grid point
            VertexStepSize =                5.0 * I3Units.m,        # ! Use small vertex size
            ZenithReach =                   2.0 * I3Units.degree,   # ! Small reach for icecube
            AzimuthReach =                  2.0 * I3Units.degree,   # ! Small reach for icecube
            GridpointVertexCorrection =     name+"_GridpointVertexCorrection", # ! Name of vertex correction service
            Minimizer =                     MinimizerName,        # minimizer service has been installed earlier
            NumberOfSamplingPoints =        8,                      # Default
            NumberOfSteps =                 3,                      # Default
            MCTruthName =                   "",                     # Default
            If = If
            )


@icetray.traysegment
def OnlineL2Filter(tray, name, pulses='CleanedMuonPulses', llhfit_name='PoleMuonLlhFit',
        angular_error=False, spline=False,
        PhotonicsTabledirMu = None, PhotonicsDriverdirMu_Spice1 = None, PhotonicsDriverfileMu_Spice1 = None,
        SplineDir = None,
        If = lambda f: True,):
    """
    This tray segment is doing the 2014 Online Level 2 processing and filtering.

    :param pulses:
        Name of the I3RecoPulseSeriesMap to work on.  Use those coming out of the BaseProcessing
        segment (filter_globals.CleanedMuonPulses)
    :param llhfit_name:
        Name of the I3Particle containing the muon llh reconstruction result from the muon
        filter BaseProcessing segment.
    :param angular_error:
        Do more advanced angular error estimation (takes more processing time)
    :param spline:
        Do the SplineMPE reco after Online L2 cuts
    :param PhotonicsTabledirMu:
        Path to the directory containing photonics driver files and
        subdirectory with photonics tables (used by I3TruncatedEnergy)
    :param PhotonicsDriverfileMu_Spice1:
        Name of photonics driver file inside PhotonicsTabledirMu (used by
        I3TruncatedEnergy)
    :param SplineDir:
        Path to the photospline tables used by spline-reco
    :param If:
        Python function to use as conditional execution test for segment modules.
    """

    from os.path import expandvars
    from math import cos, radians
    from icecube import gulliver, lilliput
    # using the CommonVariables: http://icecube.wisc.edu/~mwolf/docs/projects/CommonVariables/
    from icecube.common_variables import direct_hits
    from icecube.common_variables import hit_multiplicity
    from icecube.common_variables import hit_statistics
    from icecube.common_variables import track_characteristics
    from icecube import improvedLinefit
    from icecube import filterscripts
    import filter_globals

    # load needed libs, the "False" suppresses any "Loading..." messages
    icetray.load("bayesian-priors", False)
    icetray.load("gulliver-modules", False)
    icetray.load("mue", False)
    icetray.load("cramer-rao", False)
    icetray.load("double-muon", False)
    icetray.load("photonics-service", False)
    icetray.load("gulliver", False)
    icetray.load("truncated_energy", False)

    domain = getDomain()
    #if not PhotonicsTabledirMu:
    #    # try to guess where to get photonics
    #    if domain == Domain.SPS or domain == Domain.SPTS:
    #        PhotonicsTabledirMu = "/mnt/data/pnflocal/photorec2"
    #    elif domain == Domain.WISC:
    #        PhotonicsTabledirMu = "/net/user/i3filter/photonics/photorec/mu/" # UW Madison
    #    elif domain == Domain.DESY:
    #        PhotonicsTabledirMu = "/lustre/fs6/group/i3/mvoge/photonics/photorec/mu/" # DESY Zeuthen
    #    else:
    #        PhotonicsTabledirMu = "/opt/photorec2"
    PhotonicsTabledirMu_Spice1 = PhotonicsTabledirMu
    #PhotonicsDriverdirMu_Spice1 = PhotonicsTabledirMu
    PhotonicsDriverdirMu_Spice1 = PhotonicsDriverdirMu_Spice1
    #if not PhotonicsDriverfileMu_Spice1:
    #    PhotonicsDriverfileMu_Spice1 = "SPICE1_photorec_mureco_z80_a20_level2_showers.list"

    #############################
    #  Filter functions
    #############################

    # Make sure the events passed the muon filter
    def ThisEventPassedMuonFilter(frame):
        returnVal = False
        if frame.Has(filter_globals.MuonFilter):
            returnVal = frame[filter_globals.MuonFilter].value
        return returnVal

    # Conditional execution for L2 events
    def ThisEventPassedOnlineL2Filter(frame):
        returnVal = False
        if frame.Has(filter_globals.OnlineL2Filter):
            returnVal = frame[filter_globals.OnlineL2Filter].value
        return returnVal

    def ThisEventIsUpGoingTrack(frame):
        minZenith = 80.*I3Units.degree
        maxZenith = 180.*I3Units.degree
        particleName = name+'_BestFit' # This is the best available fit that succeeded
        returnVal = False
        if frame.Has(particleName):
            if minZenith <= frame[particleName].dir.zenith and frame[particleName].dir.zenith <= maxZenith:
                returnVal = True
        return returnVal

    def ThisEventIsHighNChannel(frame):
        nchCut = 500
        hitMultName = name+'_HitMultiplicityValues'
        returnVal = False
        if frame.Has(hitMultName):
            nch = frame[hitMultName].n_hit_doms
            returnVal = nch >= nchCut
        return returnVal

    def ThisEventIsL2UpGoingTrack(frame):
        return ThisEventPassedOnlineL2Filter(frame) and ThisEventIsUpGoingTrack(frame)

    def DoTheSplitting(frame):
        return ThisEventPassedOnlineL2Filter(frame)

    def DoAngularErrorEstimation(frame):
        #return ThisEventPassedOnlineL2Filter(frame) and (ThisEventIsUpGoingTrack(frame) or ThisEventIsHighNChannel(frame))
        return ThisEventPassedOnlineL2Filter(frame) and ThisEventIsUpGoingTrack(frame)


    #############################
    #  Online L2 reco services
    #############################

    # Bayesian weight for downgoing muons
    tray.AddService("I3PowExpZenithWeightServiceFactory", name+"_ZenithWeight",
            Amplitude = 2.49655e-07,                  # Default
            CosZenithRange = [0.,1.],                 # Bugfix: Changed from [-1,1] (Default) to [0,1], so that outside of that, the penalty is calculated
            DefaultWeight = 1.383896526736738e-87,    # Default
            ExponentFactor = 0.778393,                # Default
            FlipTrack = False,                        # Default
            PenaltySlope = -1000.,                    # Bugfix: Changed from 0. (Default) to -1000. for having a penalty when being in wrong region => force downgoing tracks
            PenaltyValue = -200.,                     # Default
            Power = 1.67721,                          # Default
            )

    llhname_spe1st  = lilliput.__add_pandel__(tray, pulses, "SPE1st")

    # Combine pandel and downgoing muon likelihood, used for bayesian fit
    tray.AddService("I3EventLogLikelihoodCombinerFactory", name+"_ZenithWeightPandel",
            InputLogLikelihoods = [llhname_spe1st, name+"_ZenithWeight"],
            )

    llhname_mpe  = lilliput.__add_pandel__(tray, pulses, "MPE")

    mininame = lilliput.__add_simplexminimizer__(tray)

    paraname = lilliput.__add_simple_track_parametrization__(tray)






    ############################
    # First module comes here
    ############################
    # These are variable calculations for OnlineL2 filter

    # used for the following direct hit calcualtions
    dh_defs = direct_hits.default_definitions # default IceCube direct hits definitions (A, B, C, D)

    # calculate NDir, LDir etc. with CommonVariables:
    keyDictList =\
    tray.AddSegment(direct_hits.I3DirectHitsCalculatorSegment, name+'_MuonLLHDirectHits',
            DirectHitsDefinitionSeries       = dh_defs,
            PulseSeriesMapName               = pulses,
            ParticleName                     = llhfit_name,
            OutputI3DirectHitsValuesBaseName = llhfit_name+'DirectHits',
            BookIt                           = True,
            If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
            )
    llhDirectHitsKeeps = [ element['key'] for element in keyDictList ]
    filter_globals.onlinel2filter_keeps += llhDirectHitsKeeps

    # single seed for iterative llh reco; muonllh first guess
    tray.AddService("I3BasicSeedServiceFactory", name+"_multitseedprep",
            AddAlternatives = "None",
            FirstGuesses =  [llhfit_name],
            ChargeFraction = 0.9,
            MaxMeanTimeResidual = 1000*I3Units.ns,
            NChEnergyGuessPolynomial = [],
            OnlyAlternatives = False,
            SpeedPolice = True,
            InputReadout =  pulses,
            TimeShiftType = "TFirst"
            )

    tray.AddModule("I3IterativeFitter", name+"_SPE2itFit",
            RandomService = "SOBOL",
            SeedService = name+"_multitseedprep",# This service seeds SPE with PoleMuonLlhFit as FirstGuess.
            Parametrization = paraname,
            LogLikelihood = llhname_spe1st,
            NIterations = 2,
            Minimizer = mininame,
            CosZenithRange = [-1,1],
            If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
            )

    # calculate NDir, LDir etc. with CommonVariables:
    keyDictList =\
    tray.AddSegment(direct_hits.I3DirectHitsCalculatorSegment, name+'_SPE2itFit_DirectHits',
            DirectHitsDefinitionSeries       = dh_defs,
            PulseSeriesMapName               = pulses,
            ParticleName                     = name+'_SPE2itFit',
            OutputI3DirectHitsValuesBaseName = name+'_SPE2itFit'+'DirectHits',
            BookIt                           = True,
            If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
            )
    speDirectHitsKeeps = [ element['key'] for element in keyDictList ]
    filter_globals.onlinel2filter_keeps += speDirectHitsKeeps

    tray.AddService("I3BasicSeedServiceFactory", name+"_mpeseedprep", #Seed for MPE is SPE 2 iterations
            AddAlternatives = "None",
            FirstGuesses =  [name+"_SPE2itFit"],
            InputReadout =  pulses,
            MaxMeanTimeResidual = 1000*I3Units.ns,
            NChEnergyGuessPolynomial = [],
            OnlyAlternatives = False,
            SpeedPolice = True,
            TimeShiftType =  "TNone",
            )

    tray.AddModule("I3SimpleFitter", name+"_MPEFit",
            SeedService =                  name+"_mpeseedprep",
            Parametrization =              paraname,
            LogLikelihood =                llhname_mpe,
            Minimizer =                    mininame,
            If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
            )

    # calculate NDir, LDir etc. with CommonVariables:
    keyDictList = \
    tray.AddSegment(direct_hits.I3DirectHitsCalculatorSegment, name+'_MPEFit_DirectHits',
            DirectHitsDefinitionSeries       = dh_defs,
            PulseSeriesMapName               = pulses,
            ParticleName                     = name+'_MPEFit',
            OutputI3DirectHitsValuesBaseName = name+'_MPEFit'+'DirectHits',
            BookIt                           = True,
            If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
            )
    mpeDirectHitsKeeps = [ element['key'] for element in keyDictList ]
    filter_globals.onlinel2filter_keeps += mpeDirectHitsKeeps

    def findBestSucceededFit(frame):
        if If(frame) and ThisEventPassedMuonFilter(frame):
            if frame.Has(name+'_MPEFit') and frame.Has(name+'_SPE2itFit') and \
                    frame.Has(llhfit_name):
                mpefit = frame[name+'_MPEFit']
                spefit = frame[name+'_SPE2itFit']
                llhfit = frame[llhfit_name]
                if mpefit.fit_status == dataclasses.I3Particle.FitStatus.OK:
                    fitparams = frame[name+'_MPEFitFitParams']
                    frame.Put(name+"_BestFit", mpefit)
                    frame.Put(name+"_BestFitFitParams", fitparams)
                    frame.Put(name+"_BestFitName", dataclasses.I3String(name+'_MPEFit'))
                elif spefit.fit_status == dataclasses.I3Particle.FitStatus.OK:
                    fitparams = frame[name+'_SPE2itFitFitParams']
                    frame.Put(name+"_BestFit", spefit)
                    frame.Put(name+"_BestFitFitParams", fitparams)
                    frame.Put(name+"_BestFitName", dataclasses.I3String(name+'_SPE2itFit'))
                elif llhfit.fit_status == dataclasses.I3Particle.FitStatus.OK:
                    fitparams = frame[llhfit_name+'FitParams']
                    frame.Put(name+"_BestFit", llhfit)
                    frame.Put(name+"_BestFitFitParams", fitparams)
                    frame.Put(name+"_BestFitName", dataclasses.I3String(llhfit_name))
                else:
                    frame.Put(name+"_BestFitName", "None")
        return True
    tray.AddModule(findBestSucceededFit, name+"_findBestSucceededFit")

    # calculate NDir, LDir etc. with CommonVariables:
    keyDictList = \
    tray.AddSegment(direct_hits.I3DirectHitsCalculatorSegment, name+'_BestFit_DirectHits',
            DirectHitsDefinitionSeries       = dh_defs,
            PulseSeriesMapName               = pulses,
            ParticleName                     = name+'_BestFit',
            OutputI3DirectHitsValuesBaseName = name+'_BestFit'+'DirectHits',
            BookIt                           = True,
            If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
            )
    bestFitDirectHitsKeeps = [ element['key'] for element in keyDictList ]
    filter_globals.onlinel2filter_keeps += bestFitDirectHitsKeeps

    # calculate NCh etc. with CommonVariables:
    keyDictList = \
    tray.AddSegment(hit_multiplicity.I3HitMultiplicityCalculatorSegment, name+'_HitMultiplicity',
            PulseSeriesMapName                = pulses,
            OutputI3HitMultiplicityValuesName = name+'_HitMultiplicityValues',
            BookIt                            = True,
            If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
            )
    hitMultiplicityKeeps = [ element['key'] for element in keyDictList ]
    filter_globals.onlinel2filter_keeps += hitMultiplicityKeeps

    # calculate QTot etc. with CommonVariables:
    keyDictList = \
    tray.AddSegment(hit_statistics.I3HitStatisticsCalculatorSegment, name+'_HitStatistics',
            PulseSeriesMapName              = pulses,
            OutputI3HitStatisticsValuesName = name+'_HitStatisticsValues',
            BookIt                          = True,
            If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
            )
    hitStatisticsKeeps = [ element['key'] for element in keyDictList ]
    filter_globals.onlinel2filter_keeps += hitStatisticsKeeps

    # this line separates Muon Filter level from Online L2 Filter level
    ############################################################################################################################################################
    ### add your IC86 2014 online level2 filter here
    ############################################################################################################################################################
    tray.AddModule("I3FilterModule<I3OnlineL2Filter_13>", name+"_OnlineL2Filter2013",
        # Configure:
        PriParticleKey = name+'_BestFit',
        DirectHitValues = name+'_BestFitDirectHitsC',
        HitMultiplicityValues = name+'_HitMultiplicityValues',
        HitStatisticsValues = name+'_HitStatisticsValues',
        LLHFitParamsKey = llhfit_name+'FitParams',
        CosZenithZone1 = [-1., cos(radians(82.))],
        CosZenithZone2 = [cos(radians(82.)), cos(radians(66.))],
        CosZenithZone3 = [cos(radians(66.)), 1.],
        # cut values:
        LDirCZone1 = 160.,
        NDirCZone1 = 9,
        PLogLParamZone1 = 4.5,
        PLogLZone1 = 8.3,
        QTotZone1 = 2.7,
        QTotSlopeZone2 = 3.3,
        QTotInterceptZone2 = 1.05 + 0.08,
        PLogLParamZone2 = 4.5,
        PLogLZone2 = 8.3,
        QTotOrCutZone2 = 2.5,
        QTotKinkZone3 = 0.5,
        QTotSlopeZone3 = 0.6,
        QTotInterceptZone3 = 2.4 + 0.13,
        PLogLParamZone3 = 4.5,
        PLogLZone3 = 10.5,
        QTotOrCutZone3 = 3.,
        DiscardEvents = False,
        DecisionName = filter_globals.OnlineL2Filter,
        If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
        )

# Don't do again, pulse mask already there offline
    #### Store a pulse mask pointing to the pulses used for reconstructions,
    #### but only for events passing OnlineL2
    ###tray.AddModule(PulseMaskShortcutter, name + "_PulseMaskShortcutter",
    ###        PulseMaskName = pulses,
    ###        ShortcutName = name+"_"+pulses,
    ###        OrigSourceName = "I3SuperDST",
    ###        If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f)
    ###        )

    # now, more high level variables are calculated

    tray.AddModule( "I3mue", name + "_BestFitMuE",
            RecoPulseSeriesNames = [pulses],
            RecoResult = name+"_BestFit", # This is the best available fit that succeeded
            RecoIntr = 1,
            OutputParticle = name+"_BestFit_MuE",
            Verbose =  0 ,
            RecoRnum =  8,
            If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f)
            )

    if domain == Domain.SPTS or domain == Domain.SPS:
        PathToTable = '/usr/local/pnf/jeb/cramer-rao/resources/data/'
    else:
        PathToTable = expandvars("$I3_SRC/cramer-rao/resources/data/")

    tray.AddModule("CramerRao", name + "_CramerRaoSPE_2it",
            InputResponse = pulses,
            InputTrack = name+"_SPE2itFit",
            OutputResult = name+"_CramerRao_SPE2itFit",
            AllHits = True,
            DoubleOutput = True,
            PathToTable = PathToTable,
            z_dependent_scatter = True,
            If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f)
            )

    tray.AddModule("CramerRao", name + "_CramerRaoMPE",
            InputResponse = pulses,
            InputTrack = name+"_MPEFit",
            OutputResult = name+"_CramerRao_MPEFit",
            AllHits = True,
            DoubleOutput = True,
            PathToTable = PathToTable,
            z_dependent_scatter = True,
            If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f)
            )

    tray.AddModule("CramerRao", name + "_CramerRaoBestFit",
            InputResponse = pulses,
            InputTrack = name+"_BestFit",
            OutputResult = name+"_CramerRao_BestFit",
            AllHits = True,
            DoubleOutput = True,
            PathToTable = PathToTable,
            z_dependent_scatter = True,
            If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f)
            )



    tray.AddService("I3BasicSeedServiceFactory", name + "_BayesianSeedPrep",
            InputReadout =  pulses,
            TimeShiftType =  "TFirst",
            FirstGuesses =  [name+"_BestFit"], # This is the best available fit that succeeded
            AddAlternatives = "None",
            AltTimeShiftType = "TFirst",
            ChargeFraction = 0.9,
            NChEnergyGuessPolynomial = [],
            OnlyAlternatives = False,
            SpeedPolice = True,
            )

    tray.AddModule("I3SimpleFitter", name + "_BayesianFit",
            SeedService = name+"_BayesianSeedPrep",
            Parametrization = paraname,
            LogLikelihood = name+"_ZenithWeightPandel",
            Minimizer = mininame,
            If = lambda f: If(f) and ThisEventIsL2UpGoingTrack(f)
            )


    # create PulseMap with only the First Pulse and the total charge
    tray.AddModule("I3FirstPulsifier", name+"_first-pulsify_qtot",
            InputPulseSeriesMapName = pulses,
            OutputPulseSeriesMapName = "FirstPulseMuonPulses",
            KeepOnlyFirstCharge = False,   # default
            UseMask = False,               # default
            If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
            )

    ######################
    # Split Fits start
    ######################

    # Split the pulses - TIME
    tray.AddModule("I3ResponseMapSplitter", name+"_TWCleanSplitTime",
            InputPulseMap =      "FirstPulseMuonPulses" , # ! Use timewindow cleaned pulses and only the first Pulses
            InputTrackName =     "" ,                       # Default, split by average pulse time
            MinimumNch =         2 ,                        # Default
            DoTRes =             False ,                    # Default
            MinTRes =            0.0 ,                      # Default
            MaxTRes =            1000.0 ,                   # Default
            DoBrightSt =         False ,                    # Default
            MaxDBrightSt =       150.0 ,                    # Default
            If = lambda f: If(f) and DoTheSplitting(f)
            )

    llhname_timesplit = []
    for i in (1, 2):
        # Linefit first track
        tray.AddSegment(improvedLinefit.simple, name+"_improvedLineFitTimeSplit%i"%i,
                inputResponse = name+"_TWCleanSplitTime%i"%i,
                fitName = name+"_SplitTime%i_Linefit"%i,
                If = lambda f: If(f) and DoTheSplitting(f)
                )

        llhname_timesplit += [lilliput.__add_pandel__(tray, name+"_TWCleanSplitTime%i"%i, "SPE1st")]

        tray.AddService("I3BasicSeedServiceFactory", name+"_SeedTimeSplit%i"%i,
                FirstGuesses =              [ name+"_SplitTime%i_Linefit"%i ], # ! Use linefit first guess
                InputReadout =              name+"_TWCleanSplitTime%i"%i,     # ! Use pulses for
                TimeShiftType =             "TFirst",                 # ! Use TFirst
                ChargeFraction =            0.9,                      # Default
                FixedEnergy =               float( "nan" ),           # Default
                MaxMeanTimeResidual =       1000.0*I3Units.ns,        # Default
                NChEnergyGuessPolynomial =  [],                       # Default
                SpeedPolice =               True,                     # Default
                AddAlternatives =           "None",                   # Default
                AltTimeShiftType =          "TFirst",                 # Default
                OnlyAlternatives =          False                     # Default
                )

        tray.AddModule("I3IterativeFitter", name+"_SplitTime%i_SPE2itFit"%i,
                RandomService =      "SOBOL",                   # ! Name of randomizer service
                NIterations =        2,                         # ! Nunmber of iterations
                SeedService =        name+"_SeedTimeSplit%i"%i,       # ! Name of seed service
                Parametrization =    paraname,             # ! Name of track param service
                LogLikelihood =      llhname_timesplit[i-1],     # ! Name of likelihood service
                CosZenithRange =     [ -1, 1 ],                # Default
                Minimizer =          mininame,                  # ! Name of minimizer service
                If = lambda f: If(f) and DoTheSplitting(f)
                )

        tray.AddService("I3BasicSeedServiceFactory", name+"_BayesianSeedPrep_TimeSplit%i"%i,
                InputReadout =  name+"_TWCleanSplitTime%i"%i,
                TimeShiftType =  "TFirst",
                FirstGuesses =  [name+"_SplitTime%i_SPE2itFit"%i],
                AddAlternatives = "None",
                AltTimeShiftType = "TFirst",
                ChargeFraction = 0.9,
                NChEnergyGuessPolynomial = [],
                OnlyAlternatives = False,
                SpeedPolice = True,
                )

        tray.AddModule("I3SimpleFitter", name+"_SplitTime%i_BayesianFit"%i,
                SeedService = name+"_BayesianSeedPrep_TimeSplit%i"%i,
                Parametrization = paraname,
                LogLikelihood = name+"_ZenithWeightPandel",
                Minimizer = mininame,
                If = lambda f: If(f) and DoTheSplitting(f) and ThisEventIsL2UpGoingTrack(f)
                )
        #end loop



    # Split the pulses - GEO
    tray.AddModule("I3ResponseMapSplitter", name+"_TWCleanSplitGeo",
            InputPulseMap =     "FirstPulseMuonPulses",  # ! Use timewindow cleaned pulses and only the first Pulses
            InputTrackName =    name+"_BestFit",               # This is the best available fit that succeeded
            MinimumNch =        2 ,                        # Default
            DoTRes =            False ,                    # Default
            MinTRes =           0.0 ,                      # Default
            MaxTRes =           1000.0 ,                   # Default
            DoBrightSt =        False ,                    # Default
            MaxDBrightSt =      150.0 ,                    # Default
            If = lambda f: If(f) and DoTheSplitting(f)
            )

    llhname_geosplit = []
    for i in (1, 2):
        # Linefit first track
        tray.AddSegment(improvedLinefit.simple, name+"_improvedLineFitGeoSplit%i"%i,
                inputResponse = name+"_TWCleanSplitGeo%i"%i,
                fitName = name+"_SplitGeo%i_Linefit"%i,
                If = lambda f: If(f) and DoTheSplitting(f)
                )

        llhname_geosplit += [lilliput.__add_pandel__(tray, name+"_TWCleanSplitGeo%i"%i, "SPE1st")]

        tray.AddService("I3BasicSeedServiceFactory", name+"_SeedGeoSplit%i"%i,
                FirstGuesses =              [ name+"_SplitGeo%i_Linefit"%i ], # ! Use linefit first guess
                InputReadout =              name+"_TWCleanSplitGeo%i"%i,     # ! Use pulses
                TimeShiftType =             "TFirst",                  # ! Use TFirst for vertex correction
                ChargeFraction =            0.9,                       # Default
                FixedEnergy =               float( "nan"  ),           # Default
                MaxMeanTimeResidual =       1000.0*I3Units.ns,         # Default
                NChEnergyGuessPolynomial =  [],                        # Default
                SpeedPolice =               True ,                     # Default
                AddAlternatives =           "None",                    # Default
                AltTimeShiftType =          "TFirst",                  # Default
                OnlyAlternatives =          False                      # Default
                )

        # Iterative fit with two iterations
        tray.AddModule("I3IterativeFitter", name+"_SplitGeo%i_SPE2itFit"%i,
                RandomService =      "SOBOL",                  # ! Name of randomizer service
                NIterations =        2 ,                       # ! Nunmber of iterations
                SeedService =        name+"_SeedGeoSplit%i"%i,       # ! Name of seed service
                Parametrization =    paraname,            # ! Name of track service
                LogLikelihood =      llhname_geosplit[i-1],     # ! Name of likelihood service
                CosZenithRange =     [ -1, 1 ] ,               # ! This is a downgoing hypothesis
                Minimizer =          mininame,                 # ! Name of minimizer service
                If = lambda f: If(f) and DoTheSplitting(f)
                )

        tray.AddService("I3BasicSeedServiceFactory", name+"_BayesianSeedPrep_GeoSplit%i"%i,
                InputReadout =  name+"_TWCleanSplitGeo%i"%i,
                TimeShiftType =  "TFirst",
                FirstGuesses =  [name+"_SplitGeo%i_SPE2itFit"%i],
                AddAlternatives = "None",
                AltTimeShiftType = "TFirst",
                ChargeFraction = 0.9,
                NChEnergyGuessPolynomial = [],
                OnlyAlternatives = False,
                SpeedPolice = True,
                )

        tray.AddModule("I3SimpleFitter", name+"_SplitGeo%i_BayesianFit"%i,
                SeedService = name+"_BayesianSeedPrep_GeoSplit%i"%i,
                Parametrization = paraname,
                LogLikelihood = name+"_ZenithWeightPandel",
                Minimizer = mininame,
                If = lambda f: If(f) and DoTheSplitting(f) and ThisEventIsL2UpGoingTrack(f)
                )
        #end loop

    ######################
    # Split Fits stop
    ######################

    tray.AddModule("I3FirstPulsifier", name+'_first-pulsify',
            InputPulseSeriesMapName = pulses,
            OutputPulseSeriesMapName = 'FirstPulseMask',
            KeepOnlyFirstCharge = True,
            UseMask = True,                # only mask on first pulse carrying first charge
            If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f)
           )
    # calculate LEmpty and Separation with CommonVariables:
    keyDictList = \
    tray.AddSegment(track_characteristics.I3TrackCharacteristicsCalculatorSegment, name+'_TrackCharacteristics',
            PulseSeriesMapName                     = 'FirstPulseMask', # Sirin calculated Separation with first pulses only,
                                                                             # I3SeparationCalculator converts internally to first pulses, but CV not
            ParticleName                           = name+'_BestFit', # This is the best available fit that succeeded
            OutputI3TrackCharacteristicsValuesName = name+'_BestFit'+'Characteristics',
            TrackCylinderRadius                    = 150.*I3Units.m,
            BookIt                                 = True,
            If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f)
            )
    keepList = [ element['key'] for element in keyDictList ]
    filter_globals.onlinel2filter_keeps += keepList
    # calculate SAll with CommonVariables:
    keyDictList = \
    tray.AddSegment(track_characteristics.I3TrackCharacteristicsCalculatorSegment, name+'_TrackCharacteristicsNoCylinderCut',
            PulseSeriesMapName                     = 'FirstPulseMuonPulses',
            ParticleName                           = name+'_BestFit', # This is the best available fit that succeeded
            OutputI3TrackCharacteristicsValuesName = name+'_BestFit'+'CharacteristicsWithSAll',
            TrackCylinderRadius                    = 4000.*I3Units.m, # effectively turn off cylinder cut to get SAll as defined in I3CutsModule
            BookIt                                 = True,
            If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f)
            )
    keepList = [ element['key'] for element in keyDictList ]
    filter_globals.onlinel2filter_keeps += keepList


    # apply the new muex to the selected events
    # this module has not been fully vetted yet but we'll go with it anyway
    # should not take too much time and it might turn out useful
    tray.AddModule("muex", name+"_MuEx",
            pulses =                        pulses,
            rectrk =                        name+"_BestFit", # This is the best available fit that succeeded
            result =                        name+"_BestFit_MuEx",
            detail =                        False,   #(wiki:IC79_Ereco -> does not look very promising yet)
            lcspan =                        0, # We use SLC pulses
            If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f)
            )

    tray.AddService("I3PhotonicsServiceFactory", name+"_PhotonicsServiceMu_Spice1",
            PhotonicsTopLevelDirectory =     PhotonicsTabledirMu_Spice1,                # ! Path to the directory containing subdirectory with tables
            DriverFileDirectory =            PhotonicsDriverdirMu_Spice1,               # ! Directory where the driver files are located
            PhotonicsLevel2DriverFile =      PhotonicsDriverfileMu_Spice1,              # ! Name of level 2 driver file
            PhotonicsInterpolationMode =     255,                     # Default
            PhotonicsVerbosity =             0,                       # Default
            PhotonicsAngularSelectionLow =   0 * I3Units.degree,      # Default
            PhotonicsAngularSelectionHigh =  180 * I3Units.degree,    # Default
            PhotonicsZSelectionLow =         -10000* I3Units.m,       # Default
            PhotonicsZSelectionHigh =        10000 * I3Units.m,       # Default
            PhotonicsTableSelection =        2,                       # ! Use level2 tables
            ServiceName =                    name+"_PhotonicsServiceMu_Spice1",
            )

    bad_doms = [ icetray.OMKey(1,45), icetray.OMKey(1,46), icetray.OMKey(5,5),
    icetray.OMKey(5,6), icetray.OMKey(5,21), icetray.OMKey(5,22), icetray.OMKey(7,35),
    icetray.OMKey(7,36), icetray.OMKey(9,15), icetray.OMKey(11,1), icetray.OMKey(11,2),
    icetray.OMKey(18,45), icetray.OMKey(19,59), icetray.OMKey(19,60), icetray.OMKey(22,49),
    icetray.OMKey(22,50), icetray.OMKey(27,7), icetray.OMKey(27,8), icetray.OMKey(28,41),
    icetray.OMKey(28,42), icetray.OMKey(33,5), icetray.OMKey(33,6), icetray.OMKey(39,22),
    icetray.OMKey(40,51), icetray.OMKey(40,52), icetray.OMKey(44,25), icetray.OMKey(44,26),
    icetray.OMKey(44,47), icetray.OMKey(44,48), icetray.OMKey(45,19), icetray.OMKey(45,20),
    icetray.OMKey(47,55), icetray.OMKey(47,56), icetray.OMKey(49,55), icetray.OMKey(49,56),
    icetray.OMKey(53,17), icetray.OMKey(53,18), icetray.OMKey(54,51), icetray.OMKey(54,52),
    icetray.OMKey(59,45), icetray.OMKey(59,46), icetray.OMKey(62,51), icetray.OMKey(62,52),
    icetray.OMKey(66,33), icetray.OMKey(66,34), icetray.OMKey(66,45), icetray.OMKey(66,46),
    icetray.OMKey(68,42), icetray.OMKey(69,23), icetray.OMKey(69,24), icetray.OMKey(6,11),
    icetray.OMKey(7,34), icetray.OMKey(7,37), icetray.OMKey(7,38), icetray.OMKey(7,39),
    icetray.OMKey(7,40), icetray.OMKey(7,41), icetray.OMKey(7,42), icetray.OMKey(7,44),
    icetray.OMKey(7,45), icetray.OMKey(7,46), icetray.OMKey(8,59), icetray.OMKey(8,60),
    icetray.OMKey(9,16), icetray.OMKey(18,46), icetray.OMKey(24,56), icetray.OMKey(26,46),
    icetray.OMKey(29,59), icetray.OMKey(29,60), icetray.OMKey(30,23), icetray.OMKey(30,60),
    icetray.OMKey(32,57), icetray.OMKey(32,58), icetray.OMKey(34,11), icetray.OMKey(34,12),
    icetray.OMKey(34,13), icetray.OMKey(34,14), icetray.OMKey(34,15), icetray.OMKey(34,17),
    icetray.OMKey(34,22), icetray.OMKey(38,59), icetray.OMKey(39,22), icetray.OMKey(39,61),
    icetray.OMKey(42,47), icetray.OMKey(42,48), icetray.OMKey(50,36), icetray.OMKey(54,47),
    icetray.OMKey(69,44), icetray.OMKey(69,47), icetray.OMKey(69,48), icetray.OMKey(71,39),
    icetray.OMKey(74,9), icetray.OMKey(79,55), icetray.OMKey(79,56), icetray.OMKey(86,27),
    icetray.OMKey(86,28) ]

    def putBadDomList(frame):
        if If(frame) and ThisEventPassedOnlineL2Filter(frame):
            frame.Put(name+"_BadDomList", dataclasses.I3VectorOMKey(bad_doms))
    tray.AddModule(putBadDomList, name+"_BadDOMsToFrame")

    # we also apply truncated mean, it looks quite good on wiki:IC79_Ereco
    tray.AddModule("I3TruncatedEnergy", name+"_TruncatedEnergy",
            RecoPulsesName =               pulses,
            #        RecoPulsesName = "FirstPulseMuonPulses", # Need to test if only first pulses should be used
            RecoParticleName =             name+"_BestFit", # This is the best available fit that succeeded
            ResultParticleName =           name+"_BestFit_TruncatedEnergy",
            I3PhotonicsServiceName =       name+"_PhotonicsServiceMu_Spice1",   # SPICE 1 tables
            UseRDE =                       True,                        # default- Correct for HQE DOMs
            BadDomListName =               name+"_BadDomList",
            If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f)
            )

#######################################
## add additional reco modules here
#######################################

    if spline:
        tray.AddSegment(SplineReco, name,
                pulses=pulses, SplineDir=SplineDir,
                ParametrizationName=paraname, MinimizerName=mininame, PathToCramerRaoTable=PathToTable,
                If = lambda f: If(f) and ThisEventPassedOnlineL2Filter(f))

    if angular_error:
        tray.AddSegment(AngularError, name,
                pulses=pulses, MinimizerName=mininame, MPEllhName=llhname_mpe,
                If = lambda f: If(f) and DoAngularErrorEstimation(f)) # run only on upgoing

#######################################
## clean up
#######################################

    leftovers = [
                 name+"_SPE2itFit",
                 name+"_SPE2itFitFitParams",
                 name+"_MPEFit",
                 name+"_MPEFitFitParams",
                 name+"_BestFit",
                 name+"_BestFitFitParams",
                 name+"_BestFitName",
                ] + llhDirectHitsKeeps + speDirectHitsKeeps + mpeDirectHitsKeeps + \
                        bestFitDirectHitsKeeps + hitMultiplicityKeeps + hitStatisticsKeeps

    # Remove leftovers from OnlineL2 fits for events that did not pass OnlineL2
    tray.AddModule("Delete", name+"_delete_onlinel2_leftovers",
            Keys = leftovers,
            If = lambda f: If(f) and not ThisEventPassedOnlineL2Filter(f)
            )

    # Remove things we don't want to keep (online, the keeper removes them)
    tray.AddModule("Delete", name+"_delete_firstpulsemask",
            Keys = ["FirstPulseMuonPulses", "FirstPulseMask", name+"_BadDomList",
                "OnlineL2_TWCleanSplitGeo1", "OnlineL2_TWCleanSplitGeo2",
                "OnlineL2_TWCleanSplitTime1", "OnlineL2_TWCleanSplitTime2"],
            If = lambda f: If(f) and ThisEventPassedMuonFilter(f)
            )

    filter_globals.onlinel2filter_keeps += [
            name+"_"+pulses,
            name+"_SPE2itFit",  # Reco results from online L2
            name+"_SPE2itFitFitParams",
            name+"_MPEFit",
            name+"_MPEFitFitParams",
            name+"_BayesianFit",
            name+"_BayesianFitFitParams",
            name+"_CramerRao_SPE2itFitParams",
            name+"_CramerRao_SPE2itFit_cr_azimuth",
            name+"_CramerRao_SPE2itFit_cr_zenith",
            name+"_CramerRao_MPEFitParams",
            name+"_CramerRao_MPEFit_cr_azimuth",
            name+"_CramerRao_MPEFit_cr_zenith",
            name+"_CramerRao_BestFitParams",
            name+"_CramerRao_BestFit_cr_azimuth",
            name+"_CramerRao_BestFit_cr_zenith",
            name+"_SplitGeo1_Linefit",
            name+"_SplitGeo2_Linefit",
            name+"_SplitTime1_Linefit",
            name+"_SplitTime2_Linefit",
            name+"_SplitGeo1_SPE2itFit",
            name+"_SplitGeo2_SPE2itFit",
            name+"_SplitTime1_SPE2itFit",
            name+"_SplitTime2_SPE2itFit",
            name+"_SplitGeo1_SPE2itFitFitParams",
            name+"_SplitGeo2_SPE2itFitFitParams",
            name+"_SplitTime1_SPE2itFitFitParams",
            name+"_SplitTime2_SPE2itFitFitParams",
            name+"_SplitGeo1_BayesianFit",
            name+"_SplitGeo2_BayesianFit",
            name+"_SplitTime1_BayesianFit",
            name+"_SplitTime2_BayesianFit",
            name+"_SplitGeo1_BayesianFitFitParams",
            name+"_SplitGeo2_BayesianFitFitParams",
            name+"_SplitTime1_BayesianFitFitParams",
            name+"_SplitTime2_BayesianFitFitParams",
            name+"_BestFit",
            name+"_BestFitFitParams",
            name+"_BestFitName",
            name+"_BestFit_MuE",
            name+"_BestFit_MuEx",
            name+"_BestFit_TruncatedEnergy_AllBINS_MuEres",
            name+"_BestFit_TruncatedEnergy_AllBINS_Muon",
            name+"_BestFit_TruncatedEnergy_AllBINS_Neutrino",
            name+"_BestFit_TruncatedEnergy_AllDOMS_MuEres",
            name+"_BestFit_TruncatedEnergy_AllDOMS_Muon",
            name+"_BestFit_TruncatedEnergy_AllDOMS_Neutrino",
            name+"_BestFit_TruncatedEnergy_BINS_MuEres",
            name+"_BestFit_TruncatedEnergy_BINS_Muon",
            name+"_BestFit_TruncatedEnergy_BINS_Neutrino",
            name+"_BestFit_TruncatedEnergy_DOMS_MuEres",
            name+"_BestFit_TruncatedEnergy_DOMS_Muon",
            name+"_BestFit_TruncatedEnergy_DOMS_Neutrino",
            name+"_BestFit_TruncatedEnergy_ORIG_Muon",
            name+"_BestFit_TruncatedEnergy_ORIG_Neutrino",
            name+"_BestFit_TruncatedEnergy_ORIG_dEdX",
            ]
    if spline:
        filter_globals.onlinel2filter_keeps += [
                name+"_SplineMPE",
                name+"_SplineMPEFitParams",
                name+"_CramerRao_SplineMPEParams",
                name+"_CramerRao_SplineMPE_cr_azimuth",
                name+"_CramerRao_SplineMPE_cr_zenith",
                ]
    if angular_error:
        filter_globals.onlinel2filter_keeps += [
                name+"_BestFit_ParaboloidFitParams",
                ]
